unit Bugs;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, DblPointUnit, Math, ClipBrd, KLists;

const
    NO_DIRECTION = 6;                   //̐
    NormV        = 1.0 / NO_DIRECTION;  //`q̊l
    NormWeight   = 30.0;                 //̏d̕Wl
    NDelta       = 0.8;                 //ˑRψقɂ钲l
    SegWeight    = 50.0;                //􂷂̏d
    UnitMetab    = 0.1;                 //Vӗ
    MutateProbality = 0.5;              //ˑRψق̋Nm
    foodValue    = 0.3;                 //H̃GlM[
    SPEED        = 5;
(*
    //////////////////////////////////////////////////////////
        TCreature
    //////////////////////////////////////////////////////////
*)
type
    TCreature = class
        fSkinColor : TColor;
        fCoreColor : TColor;
        fAngle     : double;
        fCurPos    : TDblPoint;
        fWeight    : double;
        fRadius    : double;
        fPosHistory: array of TDblPoint;
        fStepCount : integer;
        fSpeed     : double;
        constructor Create;
        destructor  Destroy(); override;
        function mFindCreature(): TCreature;
        procedure mDraw();   virtual;
        procedure mStep();   virtual;
        procedure mReproduce();  virtual;
        procedure mMove();  virtual;
    end;
(*
    //////////////////////////////////////////////////////////
        TBacteria
    //////////////////////////////////////////////////////////
*)
type
    TBacteria = class(TCreature)
        constructor Create;
    end;
(*
    //////////////////////////////////////////////////////////
        TBug
    //////////////////////////////////////////////////////////
*)
type
    TBug = class(TCreature)
        fchrom: array[0..NO_DIRECTION - 1] of double;
        fGetFoodCount: integer;
        fFindFood: Boolean;
        constructor Create;
        procedure mMutate();
        procedure mMetabolize();
        procedure mRotate();
        function mEatable(Creature: TCreature): Boolean;
        procedure mEat(Creature: TCreature);
        procedure mMove();   override;
        procedure mStep();   override;
        procedure mReproduce();  override;
    end;

(*
    //////////////////////////////////////////////////////////
        TCreatureList
    //////////////////////////////////////////////////////////
*)
type
    TCreatureList = class(TList)
    protected
        function  Get(Index: Integer): TCreature;
        procedure Put(Index: Integer; const Value: TCreature);
        function  First: TCreature;
        function  Last: TCreature;
    public
        property Items[Index: Integer]: TCreature read Get write Put; default;
        destructor Destroy; override;
    end;
(*
    //////////////////////////////////////////////////////////
        TWorld
    //////////////////////////////////////////////////////////
*)
type
    TCreatureWorld = class
    public
        fWidth : integer;
        fHeight: integer;
        fCreatureList  : TCreatureList;
        fCanvas : TCanvas;
        fPopulation: TIntList;
        constructor Create(pWidth, pHeight: integer; pCanvas: TCanvas);
        destructor  Destroy(); override;
        function mRandomSelect(): TCreature;
        procedure mUpdate();
    end;

var
    aWorld: TCreatureWorld;

implementation

(*
    //////////////////////////////////////////////////////////
        TCreatureList
    //////////////////////////////////////////////////////////
*)
//  ----------------------------------------------------------
function TCreatureList.Get(Index: Integer): TCreature;
begin
    Result :=  TCreature( inherited Get(Index) );
end;
//  ----------------------------------------------------------
procedure TCreatureList.Put(Index: Integer; const Value: TCreature);
begin
    inherited Put( Index, Value );
end;
//  ----------------------------------------------------------
function TCreatureList.First: TCreature;
begin
    Result :=  TBug( inherited Get(0) );
end;
//  ----------------------------------------------------------
function TCreatureList.Last: TCreature;
begin
    Result :=  TBug( inherited Get(Count - 1) );
end;
//  ----------------------------------------------------------
destructor TCreatureList.Destroy;
begin
    while not Self.Count <> 0 do Self.First.Free;
    inherited Destroy;
end;
//  ----------------------------------------------------------

(*
    //////////////////////////////////////////////////////////
        TWorld
    //////////////////////////////////////////////////////////
*)
//  ----------------------------------------------------------
constructor TCreatureWorld.Create(pWidth, pHeight: integer; pCanvas: TCanvas);
begin
    inherited Create();
    fCreatureList := TCreatureList.Create;
    fWidth  := pWidth;
    fHeight := pHeight;
    fCanvas := pCanvas;
    fPopulation := TIntList.Create;
    with fCanvas do begin
        Pen.Color := clBlack;
        with Font do begin
            Color := clGray;
            Size  := 8;
            Charset := ANSI_CHARSET;
            Name    := 'Ariel';
            Style   := Style + [fsbold];
        end;
    end;
end;
//  ----------------------------------------------------------
destructor  TCreatureWorld.Destroy;
begin
    fCreatureList.Free;
    inherited;
end;
//  ----------------------------------------------------------
function TCreatureWorld.mRandomSelect(): TCreature;
begin Result := fCreatureList.Items[Random(fCreatureList.Count)]; end;
//  ----------------------------------------------------------
procedure TCreatureWorld.mUpdate();
var
    i: integer;
    j: integer;
    aCreature: TCreature;
    bCreature: TCreature;
begin
    fCanvas.Brush.Color := clWhite;
    fCanvas.FillRect(Bounds(0, 0, fWidth, fHeight));

    fPopulation.Add(fCreatureList.Count);

    for i := fCreatureList.Count - 1 downto 0 do begin
        aCreature := fCreatureList.Items[i];
        aCreature.mStep();
        aCreature.mDraw();

        if aCreature.fWeight <= 0.5 then begin
            aCreature.Free;
        end else if aCreature.fWeight > SegWeight then begin
            aCreature.mReproduce();
        end;
    end;

    if Random > 0.2 then Exit;
    aCreature := TBacteria.Create;

{
    //TBug - TBacteria Ԑ
    for i := fCreatureList.Count - 1 downto 0 do begin
        aCreature := fCreatureList.Items[i];
        if aCreature.ClassType <> TBug then Continue;
        for j := fCreatureList.Count - 1 downto 0 do begin
            bCreature := fCreatureList.Items[j];
            if bCreature.ClassType = TBug then Continue;

            ConnectDblPoint(aCreature.fCurPos, bCreature.fCurPos, fCanvas);
            with CastPoint(MidPoint(aCreature.fCurPos, bCreature.fCurPos)) do
                fCanvas.TextOut(X, Y, FloatToStr(Ceil(DistanceAtoB(aCreature.fCurPos, bCreature.fCurPos))));
        end;
    end;
}
end;

(*
    //////////////////////////////////////////////////////////
        TCreature
    //////////////////////////////////////////////////////////
*)
//  ----------------------------------------------------------
constructor TCreature.Create;
begin
    inherited Create;
    fSkinColor := clBlack;
    fCoreColor := RGB(LimitRand(100, 256), LimitRand(100, 256), LimitRand(100, 256));
    fCurPos := DblPoint(LimitRand(0, aWorld.fWidth), LimitRand(0, aWorld.fHeight));
    fAngle  := Random * PI * 2;
    fStepCount := 0;
    SetLength(fPosHistory, 1);
    fPosHistory[fStepCount] := fCurPos;
    fSpeed := 0;
    aWorld.fCreatureList.Add(Self);
    fRadius := 30;
end;
//  ----------------------------------------------------------
destructor TCreature.Destroy();
begin
    aWorld.fCreatureList.Remove(Self);
    inherited Destroy;
end;
//  ----------------------------------------------------------
function TCreature.mFindCreature(): TCreature;
var
    i: integer;
    mindistance2: double;
    distance2   : double;
    aCreature: TCreature;
begin
    mindistance2 := MaxDouble;
    distance2    := MaxDouble;
    Result := nil;
    //̒Sł߂CreatureT
    for i := 0 to aWorld.fCreatureList.Count - 1 do begin
        aCreature := aWorld.fCreatureList.Items[i];
        if aCreature = Self Then Continue;
        distance2 := DistanceAtoBxx2(fCurPos, aCreature.fCurPos);
        if distance2 < mindistance2 then begin
            mindistance2 := distance2;
            Result := aCreature;
        end;
    end;
    //̒S r ȏ͂ȂĂȂp
    if mindistance2 > fRadius * fRadius then Result := nil;
end;
//  ----------------------------------------------------------
procedure TCreature.mDraw();
var
    OldBkMode:Integer;
begin
    with aWorld do begin
        fCanvas.Pen.Color   := fSkinColor;
        fCanvas.Brush.Color := fCoreColor;

        DrawCircle(fCurPos, Log2(fWeight + 1), fCanvas);

        fCanvas.Pen.Color   := fSkinColor;
        OldBkMode := SetBkMode(fCanvas.Handle, Integer(TRANSPARENT));
        with CastPoint(fCurPos) do begin
            fCanvas.TextOut(X + 5, Y, Format('%2.2f', [fWeight]));
            fCanvas.TextOut(X + 5, Y + 10, Format('%dth', [aWorld.fCreatureList.IndexOf(Self)]));
        end;
        SetBkMode(fCanvas.Handle, OldBkMode);

        fCanvas.Pen.Color := RGB(200,200,200);
        fCanvas.Brush.Style := bsClear;
        DrawCircle(fCurPos, fRadius, fCanvas);
        ConnectDblPoint(fCurPos, GetEndPoint(fCurPos, fRadius * 0.5, fAngle), fCanvas);
    end;
end;
//  ----------------------------------------------------------
procedure TCreature.mMove();
begin

    fCurPos := GetEndPoint(fCurPos, fSpeed, fAngle);
    //͂ݏo
    with fCurPos do begin
        if X > aWorld.fWidth then begin
//            X := aWorld.fWidth;
            X := X - aWorld.fWidth;
        end else if X < 0 then begin
//            X := 0;
            X := X + aWorld.fWidth;
        end;
        if Y > aWorld.fHeight then begin
//            Y := aWorld.fHeight;
            Y := Y - aWorld.fHeight;
        end else if Y < 0 then begin
//            Y := 0;
            Y := Y + aWorld.fHeight;
        end;
    end;

end;
//  ----------------------------------------------------------
procedure TCreature.mStep();
begin
    Inc(fStepCount);
    SetLength(fPosHistory, Length(fPosHistory) + 1);
    fPosHistory[fStepCount] := fCurPos;
end;
//  ----------------------------------------------------------
procedure TCreature.mReproduce();
begin
    //Stub
end;

//  ----------------------------------------------------------
(*
    //////////////////////////////////////////////////////////
        TBacteria
    //////////////////////////////////////////////////////////
*)
constructor TBacteria.Create();
begin
    inherited Create;
    fCoreColor := clYellow;
    fRadius := 0;
    fWeight := LimitRand(10, 200);
    fSpeed  := 1;
end;
//  ----------------------------------------------------------
(*
    //////////////////////////////////////////////////////////
        TBug
    //////////////////////////////////////////////////////////
*)
//  ----------------------------------------------------------
constructor TBug.Create();
var
    i: integer;
begin
    inherited Create;
    for i := Low(fchrom) to High(fchrom) do fchrom[i] := NormV;
    fCoreColor := clLime;
    fWeight := NormWeight;
    fGetFoodCount := 0;
    fSpeed := 5;
end;
//  ----------------------------------------------------------
procedure TBug.mMutate();
var
    index: integer;
    i: integer;
begin
    if Random > MutateProbality then Exit;
    index := Random(NO_DIRECTION);  //ˑRψقN`qCfbNX
    for i := Low(fchrom) to High(fchrom) do begin
        if i <> index then fchrom[i] := fchrom[i] * NDelta;
    end;
    fchrom[index] := 1.0 - (Sum(fchrom) - fchrom[index]);
end;
//  ----------------------------------------------------------
procedure TBug.mMetabolize();
begin
    fWeight := fWeight - UnitMetab;
end;
//  ----------------------------------------------------------
procedure TBug.mRotate();
var
    i, j: integer;
    rand, q: double;
begin
    rand := Random;
    q := 0.0;
    for i := 0 to NO_DIRECTION - 1 do begin
        j := i;
        if (q <= rand) and (rand < q + fchrom[i]) then Break;
        q := q + fchrom[i];
    end;
    if j >= NO_DIRECTION - 1 then j := NO_DIRECTION - 2;
    fAngle := j * 2 * PI / NO_DIRECTION;
end;
//  ----------------------------------------------------------
procedure TBug.mReproduce();
var
    wt: double;
    Child:TBug;
begin
    wt := fWeight / 2;   //̏d2ɕ
    fWeight := wt;
    mMutate();           //ˑRψ

    Child := TBug.Create;   //C
    Child.fPosHistory[0] := fCurPos;
    Child.fCurPos := fCurPos;
    Child.fchrom  := fchrom;   //eRs[
    Child.fAngle  := fAngle;
    Child.fWeight := wt;
    Child.mMutate();         //ˑRψ
end;
//  ----------------------------------------------------------
function TBug.mEatable(Creature: TCreature): Boolean;
begin
    if Creature <> nil then begin
        Result := (Creature.ClassType = TBacteria);
        fFindFood := Result;
    end else begin
        Result := False;
    end;
end;
//  ----------------------------------------------------------
procedure TBug.mEat(Creature: TCreature);
begin
    fWeight := fWeight + Creature.fWeight * foodValue;
    Creature.fWeight := 0;
    Inc(fGetFoodCount);
end;
//  ----------------------------------------------------------
procedure TBug.mMove();
begin

    mRotate();
    inherited;

end;
//  ----------------------------------------------------------
procedure TBug.mStep();
var
    aFood: TCreature;
begin
    mMove();
    mMetabolize();
    aFood := mFindCreature();
    if mEatable(aFood) then mEat(aFood);
    inherited;
end;
//  ----------------------------------------------------------
end.
